//$Author: behle $
//$Date: 2006-07-14 16:09:21 +0200 (Fri, 14 Jul 2006) $

// azove, Another Zero One Vertex Enumeration tool
// Copyright (C) 2006, Markus Behle
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


#include "conbddbuilder.hpp"
  
ADD* conBDDbuilder::buildBDD(){
  int min = constant;
  int max = constant;

  for(std::vector<int>::iterator it = nzCoeff.begin(); it != nzCoeff.end(); ++it){
    if(*it < 0) 
      min += *it;
    else 
      max += *it;
  }

  //if all coefficients are nonnegative [min == constant] the size of the computedTable
  // can be reduced to |c| which is smaller than the sum of all coefficients
  // for nonredundant [abs(constant) < (max - min)] inequalities
  const int cTsize = (min == constant && abs(constant) < (max - min)) ? abs(constant) : max - min;
  computedTable.resize(nzCoeffInd.size(), computedInnerTable(cTsize + 1));

  position = 0;
  cToffset = -min;
  localConstant = constant;
  return (sense == 'L') ? leqZero(min, max) : eqZero(min, max);
}

//-----------------------------------------------------------------------------------

//solve the Knapsack problem
ADD* conBDDbuilder::leqZero(const int min, const int max){
  if(min > 0)
    return new ADD(mgr->addZero());
  if(max <= 0)
    return new ADD(mgr->addOne());
  if(computedTable[position][cToffset + localConstant])  
    return new ADD(mgr, computedTable[position][cToffset + localConstant]);

  const int localCoeff = nzCoeff[position];

  ++position;
  const ADD* low = (localCoeff < 0) ? leqZero(min - localCoeff, max) : leqZero(min, max - localCoeff);
  localConstant += localCoeff;
  const ADD* high = (localCoeff < 0) ? leqZero(min, max + localCoeff) : leqZero(min + localCoeff, max);
  localConstant -= localCoeff;
  --position;
  
  ADD* result = new ADD(mgr->addVar(nzCoeffInd[position]).Ite(*high, *low));

  delete high;
  delete low;
  computedTable[position][cToffset + localConstant] = result->getNode();
  
  return result;
}

//-----------------------------------------------------------------------------------

//solve the Subset-Sum problem
ADD* conBDDbuilder::eqZero(const int min, const int max){
  if(min > 0 || max < 0)
    return new ADD(mgr->addZero());
  if(min == 0 && max == 0)
    return new ADD(mgr->addOne());
  if(computedTable[position][cToffset + localConstant])
    return new ADD(mgr, computedTable[position][cToffset + localConstant]);

  const int localCoeff = nzCoeff[position];

  ++position;
  const ADD* low = (localCoeff < 0) ? eqZero(min - localCoeff, max) : eqZero(min, max - localCoeff);
  localConstant += localCoeff;
  const ADD* high = (localCoeff < 0) ? eqZero(min, max + localCoeff) : eqZero(min + localCoeff, max);
  localConstant -= localCoeff;
  --position;
 
  ADD* result = new ADD(mgr->addVar(nzCoeffInd[position]).Ite(*high, *low));

  delete high;
  delete low;
  computedTable[position][cToffset + localConstant] = result->getNode();

  return result;
}
