//$Author: behle $
//$Date: 2007-03-01 17:13:04 +0100 (Thu, 01 Mar 2007) $

// azove, Another Zero One Vertex Enumeration tool
// Copyright (C) 2006, Markus Behle
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


#ifndef VEBDD_H
#define VEBDD_H

#include <stdio.h>
#include <vector>
#include <queue>

#include "cuddObj.hh"
#include "conbddbuilder.hpp"

#ifdef USE_TIMER
#include "usertimer.hpp"
#endif


enum{ //andOrder
  smallestFirst = 0,
  largestFirst = 1,
  smallestFirstRandom = 2,
  largestFirstRandom = 3
};
  
enum{ //outputNrOfPaths
 noOutput = 0,
 doubleValue = 1,
 exactArbitraryPrec = 2
};

typedef std::pair<ADD*, unsigned int> BDDnc;

struct BDDPredicate {
  const int andOrder;
  BDDPredicate(int andOrder_ = smallestFirst) : andOrder(andOrder_) {}
  bool operator()(const BDDnc& a, const BDDnc& b){
    switch(andOrder){
      //smallest BDDs first: faster but needs more memory
      case smallestFirst: return (a.second > b.second);
      
      //largest BDDs first: less memory consumption but slower
      case largestFirst: return (a.second < b.second);
      
      //smallest BDDs first and randomly if node counts equal: usually faster but needs more memory
      case smallestFirstRandom: return (a.second == b.second) ? (drand48() < 0.5) : (a.second > b.second);
      
      //largest BDDs first and randomly if node counts equal: usually less memory consumption but slower
      case largestFirstRandom: return (a.second == b.second) ? (drand48() < 0.5) : (a.second < b.second);
    }
    return false;
  };
};
      

class veBDD{
public:
  veBDD(const char* problemName, bool randomVarOrder, int outputNrOfPaths_, int andOrder_);
  ~veBDD() { if(resultBDD) delete resultBDD; };

  void buildConstraintBDDs(void);
  void andBDDs(void);
  void output01Points(void);
  
  void outputDotFile(char* filename){
    FILE* f = fopen(filename, "w");
    Cudd_DumpDot(mgr.getManager(), 1, &rootnode, NULL, NULL, f);
    fclose(f);
    //system("dot -Tps test.dot -o test.ps");
  }

private:
  void createVarOrder(bool randomVarOrder);
  
  inline void output(const unsigned int* point) const;
  
  unsigned int nrOfCols; //number of read in columns (= number of levels in the BDD)
  unsigned int nrOfRows; //number of read in rows

  int outputNrOfPaths;
  
  std::vector<unsigned int> index2level;
  std::vector<unsigned int> level2index;
    
  std::vector<conBDDbuilder> constraints;

  Cudd mgr;

  std::priority_queue<BDDnc, std::vector<BDDnc>, BDDPredicate> prQueue;
    
  ADD* resultBDD;
  
  DdNode* rootnode;
  DdNode* leaf1node;
  DdNode* leaf0node;

#ifdef USE_TIMER
  userTimer timeConBDDs;
  userTimer timeAndBDD;
  userTimer timeOutputPoints;
#endif

  DdApaNumber preciseSatCount; //number of different paths from the root to the leaf1
  int preciseSatCountDigits;  //needed to print preciseSatCount
};


#endif
